import salabim as sim
import pytest


def test_queue_iter():
    env = sim.Environment()
    x = [sim.Component(name="x.") for _ in range(20)]
    q = sim.Queue("q", fill=x[:10])
    collect = []
    for c in q:
        collect.append(c)
        if c == x[5]:
            x[2].leave(q)
            x[3].leave(q)
            x[3].enter(q)
            x[7].priority(q, -1)
            x[11].enter(q)
            x[12].enter(q)
        if c == x[12]:
            x[1].leave(q)
            x[1].enter(q)
            x[13].enter(q)

    assert collect == [x[0], x[1], x[2], x[3], x[4], x[5], x[7], x[6], x[8], x[9], x[3], x[11], x[12], x[1], x[13]]


def test_queue_reversed():
    env = sim.Environment()
    x = [sim.Component(name="x.") for _ in range(20)]
    q = sim.Queue("q", fill=x[:10])
    collect = []
    for c in reversed(q):
        collect.append(c)
        if c == x[5]:
            x[2].leave(q)
            x[3].leave(q)
            x[3].enter(q)
            x[7].priority(q, -1)
            x[11].enter(q)
            x[12].enter(q)
        if c == x[12]:
            x[1].leave(q)
            x[1].enter(q)
            x[13].enter(q)

    assert collect == [x[9], x[8], x[7], x[6], x[5], x[12], x[13], x[1], x[11], x[3], x[4], x[0]]


def test_queue_operations():
    env = sim.Environment()
    x = [sim.Component(name="x.") for _ in range(9)]
    q0 = sim.Queue("q.", fill=x[:6])
    q1 = sim.Queue("q.", fill=x[3:])

    qintersection = q0.intersection(q1)
    qand = q0 & q1
    assert qintersection == qand == x[3:6]
    qdifference = q0.difference(q1)
    qminus = q0 - q1
    assert qdifference == qminus == x[:3]
    qunion = q0.union(q1)
    qplus = q0 + q1
    qor = q0 | q1
    assert qunion == qplus == qor == sim.Queue(fill=x)
    qsymmetric_difference = q0.symmetric_difference(q1)
    qcaret = q0 ^ q1
    qcaretrev = q1 ^ q0
    assert qsymmetric_difference == qcaret == qcaretrev == x[:3] + x[6:]
    qcopy = q0.copy()
    assert qcopy == q0
    q1 = q0.move()
    assert len(q0) == 0
    assert q1 == sim.Queue(fill=x[:6])
    q0.clear()
    assert len(q0) == 0
    q0 = sim.Queue("q.", fill=x[:6])
    q1 = sim.Queue("q.", fill=x[3:])
    q0.extend(q1)
    assert q0 == x


def test_queue_comparisons():
    env = sim.Environment()
    x = [sim.Component(name="x.") for _ in range(9)]
    assert sim.Queue(fill=x[:3]) == sim.Queue(fill=x[:3]) == x[:3]
    assert sim.Queue(fill=x[:3]) == sim.Queue(fill=x[:3]) == set(x[:3])
    assert sim.Queue(fill=x[:3]) <= x
    assert sim.Queue(fill=x[:3]) < x
    assert x >= sim.Queue(fill=x[:3])
    assert x > sim.Queue(fill=x[:3])
    assert not (sim.Queue(fill=x[:3]) >= x)
    assert not (sim.Queue(fill=x[:3]) > x)
    assert not (x <= sim.Queue(fill=x[:3]))
    assert not (x < sim.Queue(fill=x[:3]))
    assert sim.Queue(fill=x[:3]) != sim.Queue(fill=x[2:3])


def test_queue_naming():
    env = sim.Environment()
    q0 = sim.Queue("q.")
    q1 = sim.Queue("q.")
    assert q0.name() == "q.0"
    assert q1.name() == "q.1"
    q0copy = q0.copy()
    assert q0copy.name() == "copy of q.0"
    q0copy.rename("q0copy")
    assert q0copy.name() == "q0copy"


def test_queue_priorities():
    env = sim.Environment()
    x = [sim.Component(name="x.") for _ in range(9)]
    q = sim.Queue("q")

    x[0].enter(q)
    x[1].enter_sorted(q, priority=1)
    x[2].enter(q)
    x[3].enter_sorted(q, priority=-1)
    x[4].enter_at_head(q)
    assert list(q) == [x[4], x[3], x[0], x[1], x[2]]
    assert [c.priority(q) for c in q] == [-1, -1, 0, 1, 1]

    # resort after priority change
    x[0].priority(q, 2)
    assert list(q) == [x[4], x[3], x[1], x[2], x[0]]
    assert [c.priority(q) for c in q] == [-1, -1, 1, 1, 2]


def test_queue_priorities_update():
    env = sim.Environment()
    x = [sim.Component(name="x.") for _ in range(9)]
    q = sim.Queue("q")

    for i, c in enumerate(x):
        c.enter_sorted(q, priority=-i)
    assert list(q) == list(reversed(x))

    for i, c in enumerate(q):
        c.priority(q, -i)
    assert list(q) == x
    
    
def test_queue_capacity():
    class X(sim.Component):
        pass
        
    q0 = sim.Queue('q0')
    q1 = sim.Queue('q1', capacity=4)
    for _ in range(3):
        x = X()
        x.enter(q0)
        x.enter(q1)
    x = X()
    q0.capacity.value = 4
    x.enter(q0)
    x.enter(q1)
    x = X()
    with pytest.raises(sim.QueueFullError):
        x.enter(q0)
    with pytest.raises(sim.QueueFullError):
        x.enter(q1) 
    q0.capacity.value=3
    assert len(q0) == 4
    q0.capacity.value = sim.inf
    x = X()
    x.enter(q0)  
    assert len(q0) == 5
    q2 = q0.copy() 
    q3 = q1.copy(copy_capacity=True)
    x = X()
    with pytest.raises(sim.QueueFullError):
        x.enter(q3)
    q4 = q1 + q1 
    assert q4.capacity.value == sim.inf  
    q5 = q1.move()
    assert q1.capacity.value == 4
    assert q5.capacity.value == sim.inf
    q6 = q1.move(copy_capacity=True)
    assert q1.capacity.value == 4
    assert q6.capacity.value == 4


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import salabim as sim
import pytest


def test_1(capsys):
    class X(sim.Component):
        def process(self):
            self.env.order.append(self)

    env = sim.Environment(trace=True)
    env.order = []
    x = [X(name="x.", process="") for _ in range(10)]
    x[0].activate()
    x[1].activate(at=10)
    x[2].activate(at=10, urgent=True)
    x[3].activate(delay=10)
    x[4].activate(at=10, urgent=True)
    x[5].activate(at=10)
    env.run()
    assert env.order == [x[0], x[4], x[2], x[1], x[3], x[5]]
    out, err = capsys.readouterr()
    assert "no events left" in out


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import pytest
import collections
from pathlib import Path
import os
import sys

if __name__ == "__main__":
    file_folder = Path(__file__).parent
    top_folder = (file_folder / "..").resolve()
    sys.path.insert(0, str(top_folder))
    os.chdir(file_folder)

import salabim as sim

def collect(dis, n=10000):
    m = sim.Monitor()
    for _ in range(n):
        m.tally(dis())
    return m


def test_randomseeds():
    env = sim.Environment()
    sim.random_seed(123)
    m123_0 = collect(sim.Uniform(1, 2), n=10)
    sim.random_seed(123)
    m123_1 = collect(sim.Uniform(1, 2), n=10)
    sim.random_seed(345)
    m345_0 = collect(sim.Uniform(1, 2), n=10)
    sim.random_seed("*")
    mr_0 = collect(sim.Uniform(1, 2), n=10)
    sim.random_seed("*")
    mr_1 = collect(sim.Uniform(1, 2), n=10)
    assert m123_0.x() == m123_1.x()
    assert m123_0.x() != m345_0.x()
    assert mr_0.x() != mr_1.x()


def test_constant():
    env = sim.Environment()
    m = collect(sim.Constant(1))
    assert m.mean() == 1


def test_uniform():
    m = collect(sim.Uniform(1, 2))
    assert m.mean() == pytest.approx(1.5, rel=1e-2)


def test_bounded():
    env = sim.Environment()
    m = collect(sim.Bounded(sim.Uniform(-2, 4), 0, 2), n=100)
    assert m.mean() == pytest.approx(1, rel=1e-1)
    assert m.minimum() >= 0
    assert m.maximum() <= 2

    m = collect(sim.Bounded(sim.Uniform(-2, 4), 5, fail_value=sim.inf), n=100)
    assert m.mean() == sim.inf

    m = sim.Monitor()
    for _ in range(100):
        m.tally(sim.Uniform(-2, 4).bounded_sample(0, 2))
    assert m.mean() == pytest.approx(1, rel=1e-1)
    assert m.minimum() >= 0
    assert m.maximum() <= 2

    m = sim.Monitor()
    for _ in range(100):
        m.tally(sim.Uniform(-2, 4).bounded_sample(5, fail_value=sim.inf))
    assert m.mean() == sim.inf

    m = collect(sim.Bounded(sim.Normal(4, 3), 0))
    assert m.mean() > 4
    assert m.minimum() >= 0


def test_triangular():
    env = sim.Environment()
    m = collect(sim.Triangular(1, 6, 2))
    assert m.mean() == pytest.approx(3, rel=1e-2)


def test_normal():
    env = sim.Environment()
    m = collect(sim.Normal(4, 3))
    assert m.mean() == pytest.approx(4, rel=1e-2)
    assert m.std() == pytest.approx(3, rel=1e-2)


def test_int_uniform():
    env = sim.Environment()
    m = collect(sim.IntUniform(1, 6))
    assert m.mean() == pytest.approx(3.5, rel=1e-2)
    assert set(m.x()) == {1, 2, 3, 4, 5, 6}
    count = collections.Counter(m.x())
    for v, n in count.items():
        assert n == pytest.approx(10000 / 6, rel=10)


def test_pdf():
    env = sim.Environment()
    d=sim.Pdf((1,10,2,70,3,20))
    assert str(d)=="Pdf"
    m = collect(sim.Pdf((1,10,2,70,3,20)))
    assert m.mean() == pytest.approx(2.1, rel=1e-2)
    m = collect(sim.Pdf((1,2,3),(10,70,20)))
    assert m.mean() == pytest.approx(2.1, rel=1e-2)
    d = {1:10, 2:70, 3:20}
    m = collect(sim.Pdf(d.keys(),d.values()))
    assert m.mean() == pytest.approx(2.1, rel=1e-2)
    m = collect(sim.Pdf(d))
    assert m.mean() == pytest.approx(2.1, rel=1e-2)


def test_pmf():
    env = sim.Environment()
    d=sim.Pmf((1,10,2,70,3,20))
    assert str(d)=="Pmf"    
    m = collect(sim.Pmf((1,10,2,70,3,20)))
    assert m.mean() == pytest.approx(2.1, rel=1e-2)
    m = collect(sim.Pmf((1,2,3),(10,70,20)))
    assert m.mean() == pytest.approx(2.1, rel=1e-2)
    d = {1:10, 2:70, 3:20}
    m = collect(sim.Pmf(d.keys(),d.values()))
    assert m.mean() == pytest.approx(2.1, rel=1e-2)
    m = collect(sim.Pmf(d))
    assert m.mean() == pytest.approx(2.1, rel=1e-2)

def test_cumpdf():
    env = sim.Environment()
    d=sim.CumPdf((1, 10, 2, 80, 3,100))
    assert str(d)=="CumPdf"
    m = collect(d)
    assert m.mean() == pytest.approx(2, rel=1e-1)

def test_cumpmf():
    env = sim.Environment()
    d=sim.CumPmf((1, 10, 2, 80, 3,100))
    assert str(d)=="CumPmf"
    m = collect(d)
    assert m.mean() == pytest.approx(2, rel=1e-1)


def test_scipy_distribution():
    try:
        import scipy.stats as st
    except ModuleNotFoundError:
        pytest.skip("could not import scipy")

    env = sim.Environment()

    m = collect(sim.External(st.norm, loc=5, scale=1))
    assert m.mean() == pytest.approx(5, rel=1e-2)

    m = collect(sim.External(st.norm, loc=1, scale=1, size=4))
    assert m.mean() == pytest.approx(1, rel=1e-2)


def test_numpy_distribution():
    try:
        import numpy as np
    except ImportError:
        pytest.skip("could not import numpy")

    env = sim.Environment(time_unit="hours")

    m = collect(sim.External(np.random.laplace, loc=5, scale=1, size=None, time_unit="days"))
    assert m.mean() == pytest.approx(env.days(5), rel=1e-2)


def test_random_distribution():
    import random
    import math

    env = sim.Environment()

    dis = sim.External(random.lognormvariate, mu=5, sigma=1)
    m = sim.Monitor()
    for _ in range(10000):
        m.tally(math.log(dis()))
    assert m.mean() == pytest.approx(5, rel=1e-2)


def test_Distribution_distribution():
    env = sim.Environment()
    m = collect(sim.Distribution("Uniform(1,2)"))
    assert m.mean() == pytest.approx(1.5, rel=1e-2)
    m = collect(sim.Distribution("U(1,2)"))
    assert m.mean() == pytest.approx(1.5, rel=1e-2)
    m = collect(sim.Distribution("UNIFORM(1,2)"))
    assert m.mean() == pytest.approx(1.5, rel=1e-2)
    with pytest.raises(NameError):
        m = collect(sim.Distribution("Unusual(1,2)"))
    with pytest.raises(NameError):
        m = collect(sim.Distribution("Uniformdistribution(1,2)"))
    with pytest.raises(AttributeError):
        m = collect(sim.Distribution("uniform(1)", time_unit="minutes"))
    env = sim.Environment(time_unit="seconds")
    m = collect(sim.Distribution('uniform(0,2, "minutes")', time_unit="hours"))
    assert m.mean() == pytest.approx(60, rel=1e-2)
    m = collect(sim.Distribution("0,2", time_unit="minutes"))
    assert m.mean() == pytest.approx(60, rel=1e-2)
    m = collect(sim.Distribution("uniform(1, time_unit='hours')", time_unit="minutes"))
    assert m.mean() == 60 * 60
    m = collect(sim.Distribution("uniform(1, time_unit='hours')"))
    assert m.mean() == 60 * 60
    m = collect(sim.Distribution("uniform(1, 1, 'hours')", time_unit="minutes"))
    assert m.mean() == 60 * 60
    m = collect(sim.Distribution("uniform(1, 1, 'hours')"))
    assert m.mean() == 60 * 60
    with pytest.raises(SyntaxError):
        m = collect(sim.Distribution("uniform(1, 1, ,'hours')", time_unit="minutes"))
    assert m.mean() == 60 * 60
    m = collect(sim.Distribution("1"))
    assert m.mean() == 1
    m = collect(sim.Distribution("1,2"))
    assert m.mean() == pytest.approx(1.5, rel=1e-2)
    m = collect(sim.Distribution("1,6,2"))
    assert m.mean() == pytest.approx(3, rel=1e-2)
    m = collect(sim.Distribution("(1,6,2)"))
    assert m.mean() == pytest.approx(3, rel=1e-2)


def test_expressions():
    c1 = sim.Uniform(1)
    c2 = sim.Uniform(2)
    c3 = sim.Uniform(3)
    c4 = sim.Uniform(4)
    dis = (c1 + c3) / c4 + 2 * c2 - (c4 - c1) * 5 / 3
    m = collect(dis, n=10)
    assert m.mean() == 0
    assert not any(x for x in m.x())  # all 0


def test_map():
    m = collect(sim.Map(sim.Uniform(-1, 1), lambda x: x if x > 0 else 0))
    assert m.minimum() == 0
    assert m.mean() == pytest.approx(0.25, rel=1e-1)
    assert m.number_of_entries() == pytest.approx(2 * m.number_of_entries(ex0=True), rel=1e-1)


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import salabim as sim
import pytest
import datetime

import collections

def test_environment():
    import dateutil.parser

    env=sim.Environment()
    env.run(1)
    assert env.now() == 1

    env=sim.Environment(datetime0=True)
    env.run(1)
    assert env.now() == 1
    assert env.time_to_str(env.now()) == "Thu 1970-01-01 00:00:01"    

    env=sim.Environment(datetime0="23-02-01")
    env.run(1)
    assert env.now() == 1
    assert env.time_to_str(env.now()) == "Wed 2023-02-01 00:00:01"   
     
    org_dateutil_parse=sim.Environment.dateutil_parse  
    sim.Environment.dateutil_parse = lambda self, spec: dateutil.parser.parse(spec, dayfirst=True, yearfirst=False) 
    env=sim.Environment(datetime0="01-02-23")
    env.run(1)
    assert env.now() == 1
    assert env.time_to_str(env.now()) == "Wed 2023-02-01 00:00:01"   

    sim.Environment.dateutil_parse = org_dateutil_parse 
    
def test_component0():
    import datetime
    class X(sim.Component):
        def process(self):
            assert env.now() == env.datetime_to_t(datetime.datetime(year=2023, month=1, day=15))
            self.hold(till="2023-02-01")
            self.hold("12:00:00")
            self.hold(datetime.timedelta(hours=12))
            assert env.now() == env.datetime_to_t(datetime.datetime(year=2023, month=2, day=2))
            self.hold(till="2023-03-01")
            assert env.now() == env.datetime_to_t(datetime.datetime(year=2023, month=3, day=1))
    env=sim.Environment(datetime0="23-01-01")
    X(at="2023-01-15")
    env.run()

def test_component1():
    class X(sim.Component):
        def process(self):
            assert env.now() == 1
            self.hold(till="20")
            self.hold("0.5")
            self.hold("0.5")
            assert env.now() == 21
            self.hold(till="30")
            assert env.now() == 30
    env=sim.Environment()
    X(at="1")
    env.run()

def test_component2():
    def sequence():
        sequence.n += 1
        return str(sequence.n)
    sequence.n=0

    class X(sim.Component):
        def process(self):
            assert env.now() == 1
            self.hold(sequence)
            assert env.now() == 2
            self.hold(sequence)
            assert env.now() == 4

    env=sim.Environment()
    with pytest.raises(ValueError):
        X(at="abc")
    X(at="1")
    env.run()

def test_no_dateutils():
    import sys
    for s in sys.modules.copy():
        if s.startswith("dateutil"):
            del sys.modules[s]

    import builtins
    orig_import = builtins.__import__

    def mock_import(*args, **kwargs):
        if args[0]=='dateutil.parser':
            raise ImportError(args[0])
        else:
            return orig_import(*args, **kwargs)

    builtins.__import__ = mock_import

    env=sim.Environment()

    with pytest.raises(ImportError):
        env=sim.Environment(datetime0="23-02-01")

    builtins.__import__ = orig_import


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import pytest
import os
import sys
from pathlib import Path

if __name__ == "__main__":
    file_folder = Path(__file__).parent
    top_folder = (file_folder / "..").resolve()
    sys.path.insert(0, str(top_folder))
    os.chdir(file_folder)

    pytest.main(["-vv", "-s"])


import salabim as sim
import pytest

def test():
    class X(sim.Component):
        def process(self):
            with pytest.raises(ValueError):
                yield self.hold(-1)
            assert env.now() == 0
            yield self.hold(-1, cap_now=True)
            assert env.now() == 0

    class Y(sim.Component):
        def process(self):
            now = env.now()
            yield self.hold(-1)
            assert env.now() == now

    env = sim.Environment(yieldless=False)    
    X()
    
    env.run(-1, cap_now=True)
    with sim.cap_now():
        Y()
        env.run(1)
            
if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import salabim as sim
import pytest


def test(capsys):
    env = sim.Environment(trace=True)
    env.run(1)
    out = capsys.readouterr()[0]
    assert "0.000 main" in out
    assert "+1.000" in out
    save_time_to_str = sim.Environment.time_to_str
    save_duration_to_str = sim.Environment.duration_to_str

    sim.Environment.time_to_str = lambda self, t: f"t={t:4.1f}"
    sim.Environment.duration_to_str = lambda self, duration: f"d={duration:4.1f}"
    env.run(1)
    out = capsys.readouterr()[0]
    assert "t= 2.0" in out
    assert "+d= 1.0" in out

    sim.Environment.time_to_str = save_time_to_str
    sim.Environment.duration_to_str = save_duration_to_str


def test_interp():
    env = sim.Environment()
    with pytest.raises(ValueError):
        sim.interp(0, [], [])
    with pytest.raises(ValueError):
        sim.interp(0, [1, 2], [3])
    assert sim.interp(5, [5], [10]) == 10
    assert sim.interp(5, [0, 10], [0, 20]) == 10
    assert sim.interp(0, [0, 10], [0, 20]) == 0
    assert sim.interp(10, [0, 10], [0, 20]) == 20
    assert sim.interp(-1, [0, 10], [0, 20]) == 0
    assert sim.interp(11, [0, 10], [0, 20]) == 20

def test_env_prefixing():
    env1 = sim.Environment()
    env = sim.Environment()
    assert sim.Component().env == env
    assert sim.Component(env=env).env == env
    assert sim.Component(env=env1).env == env1
    assert env.Component().env == env
    assert env.Component(env=env).env == env
    assert env.Component(env=env1).env == env1
    assert env1.Component().env == env1
    assert env1.Component(env=env).env == env
    assert env1.Component(env=env1).env == env1


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import salabim as sim
import pytest


def test_wait():
    class X(sim.Component):
        def process(self):
            self.hold(1)
            s1.set(1)
            self.hold(1)
            s1.set(2)
            s2.set("red")
            self.hold(2)
            s1.set(30)

    class Z1(sim.Component):
        def process(self):
            while True:
                self.wait((s1, lambda x, component, state: x / 2 > self.env.now()))
                self.hold(1.5)

    class Z2(sim.Component):
        def process(self):
            while True:
                self.wait((s2, lambda x, *_: x in ("red", "yellow")))
                self.hold(1.5)

    def red_or_yellow(x, component, state):
        return x in component.ok_colors

    class Z3(sim.Component):
        def setup(self):
            self.ok_colors = ("red", "yellow")

        def process(self):
            while True:
                self.wait((s2, red_or_yellow))
                self.hold(1.5)

    env = sim.Environment(trace=False)
    s1 = sim.State(name="s.", value=0)
    s2 = sim.State(name="s.", value="green")
    s3 = sim.State(name="s.")

    q = sim.Queue("q.")

    x = X()

    Z1()

    Z2()
    Z3()
    env.run(10)


def trace_to_events(out):
    result = []
    for l in out.split("\n"):
        try:
            t = float(l[6:17])
            c = l[18:39].rstrip()
            result.append((t, c))
        except ValueError:
            pass
    return result


def test_urgent_and_priority(capsys):
    class X(sim.Component):
        def process(self):

            if self == x[8]:
                self.hold(0, priority=-6, urgent=True)
            if self == x[7]:
                self.hold(0, priority=-6)
            pass

    class Y(sim.Component):
        def process(self):
            pass

    env = sim.Environment(trace=True)
    x = [X(priority=-i) for i in range(10)]
    y = [Y(priority=-i, urgent=True) for i in range(10)]
    x[1].cancel()
    x[1].activate(priority=-8)
    env.run(till=0, priority=-5)
    for c in x:
        print(c.name(), c.scheduled_priority())
    out = capsys.readouterr()[0]    
    events = trace_to_events(out)
    assert [event[1] for event in events] == [
        "main",
        "y.9",
        "x.9",
        "y.8",
        "x.8",
        "x.1",
        "y.7",
        "x.7",
        "x.8",
        "y.6",
        "x.6",
        "x.7",
        "y.5",
        "x.5",
        "main",
    ]
    assert [c.scheduled_priority() for c in x] == [0, None, -2, -3, -4, None, None, None, None, None]
    with pytest.raises(TypeError):
        X(process="", priority=1)

def test_value():
    env = sim.Environment(trace=False)
    s1 = sim.State(name="s.", value=0)
    env.run(1)
    s1.value.value+=1
    env.run(1)
    s1.value.value+=1
    env.run(1)

    assert s1()==2


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import pytest
from pathlib import Path
import sys
import os

if __name__ == "__main__":
    file_folder = Path(__file__).parent
    top_folder = (file_folder / "..").resolve()
    sys.path.insert(0, str(top_folder))
    os.chdir(file_folder)

import salabim as sim


class X(sim.Component):
    def setup(self, color="red"):
        self.color = color
        self.enter(components)


class Vehicle(sim.Component):
    def setup(self):
        self.enter(components)


class Car(Vehicle):
    pass


class Bus(Vehicle):
    pass


class Truck(Vehicle):
    pass


def exp(X, run_time=None, *args, **kwargs):
    global components

    env = sim.Environment()
    components = sim.Queue()
    sim.ComponentGenerator(X, *args, **kwargs)
    env.run(run_time)
    return components


def test_iat():
    components = exp(X, iat=sim.Uniform(0, 2), at=500, till=1000, force_at=True)
    assert len(components) == pytest.approx(500, rel=1e-2)
    assert components[0].enter_time(components) == 500
    assert 998 <= components[-1].enter_time(components) <= 1000

    with pytest.raises(ValueError):
        components = exp(X, iat=sim.Uniform(0, 2), at=500, till=1000, force_at=True, force_till=True)

    components = exp(X, iat=sim.Uniform(0, 2), till=1000, force_at=True)
    assert len(components) == pytest.approx(1000, rel=1e-2)
    assert components[-1].enter_time(components) <= 1000

    components = exp(X, iat=20, at=10, till=111, force_at=True)
    assert len(components) == 6
    assert components[0].enter_time(components) == 10
    assert components[-1].enter_time(components) == 110

    components = exp(X, iat=20, at=10, till=111)
    assert len(components) == 5
    assert components[-1].enter_time(components) == 110

    components = exp(X, iat=20, at=10, number=5, force_at=True)
    assert len(components) == 5
    assert components[0].enter_time(components) == 10
    assert components[-1].enter_time(components) == 90

    components = exp(X, iat=20, at=10, number=5)
    assert len(components) == 5
    assert components[0].enter_time(components) == 30
    assert components[-1].enter_time(components) == 110

    components = exp(X, run_time=90, iat=20, at=10)
    assert len(components) == 4
    assert components[0].enter_time(components) == 30
    assert components[-1].enter_time(components) == 90

    def set_a_1():
        nonlocal a
        a = 1

    a = 0
    components = exp(X, iat=20, at=10, number=5, at_end=set_a_1)
    assert a == 1


def test_spread():
    components = exp(X, at=100, till=200, number=10)
    assert len(components) == 10
    assert components[0].enter_time(components) > 100
    assert components[-1].enter_time(components) < 200

    components = exp(X, at=100, till=200, number=10, force_at=True)
    assert len(components) == 10
    assert components[0].enter_time(components) == 100
    assert components[-1].enter_time(components) < 200

    components = exp(X, at=100, till=200, number=10, force_till=True)
    assert len(components) == 10
    assert components[0].enter_time(components) > 100
    assert components[-1].enter_time(components) == 200

    components = exp(X, at=100, till=200, number=10, force_at=True, force_till=True)
    assert len(components) == 10
    assert components[0].enter_time(components) == 100
    assert components[-1].enter_time(components) == 200

    components = exp(X, at=100, till=200, number=1, force_till=True)
    assert len(components) == 1
    assert components[0].enter_time(components) == 200

    components = exp(X, at=100, till=200, number=1, force_at=True)
    assert len(components) == 1
    assert components[0].enter_time(components) == 100

    with pytest.raises(ValueError):
        components = exp(X, at=100, till=200, number=1, force_at=True, force_till=True)

    components = exp(X, at=100, till=200, number=0, force_till=True)
    assert len(components) == 0

    def set_a_1():
        nonlocal a
        a = 1

    a = 0
    components = exp(X, at=100, till=200, number=10, force_till=True, at_end=set_a_1)
    assert a == 1


def test_equidistant():
    components = exp(X, at=100, till=200, number=0, equidistant=True)
    assert len(components) == 0

    with pytest.raises(ValueError):
        components = exp(X, at=100, till=200, number=1, equidistant=True)

    with pytest.raises(ValueError):
        components = exp(X, at=300, till=200, number=3, equidistant=True)

    with pytest.raises(ValueError):
        components = exp(X, at=200, number=3, equidistant=True)

    with pytest.raises(ValueError):
        components = exp(X, iat=1, at=200, till=200, number=3, equidistant=True)

    components = exp(X, at=100, till=200, number=2, equidistant=True)
    assert len(components) == 2
    assert components[0].enter_time(components) == 100
    assert components[1].enter_time(components) == 200

    components = exp(X, at=100, till=200, number=3, equidistant=True)
    assert len(components) == 3
    assert components[0].enter_time(components) == 100
    assert components[1].enter_time(components) == 150
    assert components[2].enter_time(components) == 200

    components = exp(X, at=100, till=200, number=5, equidistant=True)
    assert len(components) == 5
    assert components[0].enter_time(components) == 100
    assert components[2].enter_time(components) == 150
    assert components[4].enter_time(components) == 200

    components = exp(X, at=100, till=100, number=3, equidistant=True)
    assert len(components) == 3
    assert components[0].enter_time(components) == 100
    assert components[1].enter_time(components) == 100
    assert components[2].enter_time(components) == 100

    def set_a_1():
        nonlocal a
        a = 1

    a = 0
    components = exp(X, at=100, till=200, number=10, equidistant=True, at_end=set_a_1)
    assert a == 1


def test_propagate():
    components = exp(X, number=1, iat=1)
    assert components[0].color == "red"
    assert components[0].name() == "x.0"

    components = exp(X, number=1, iat=1, color="blue", name="my name,")
    assert components[0].color == "blue"
    assert components[0].name() == "my name.1"


def test_dis():
    components = exp(sim.Pdf((Car, Bus, Truck), (50, 30, 20)), iat=1, number=1000)

    names = sim.Monitor()
    for component in components:
        names.tally(component.name().split(".")[0])

def test_moments():
    moments=(1,2,3,4,100,5)
    components = exp(X, moments=moments)
    assert components[0].enter_time(components) == 1
    assert components[1].enter_time(components) == 2
    assert components[2].enter_time(components) == 3
    assert components[3].enter_time(components) == 4
    assert components[4].enter_time(components) == 5
    assert components[5].enter_time(components) == 100



#    names.print_histogram(values=True, sort_on_weight=True)

if __name__ == "__main__":
    pytest.main(["-vv", "-s",__file__])





import salabim as sim
import pytest


def test_define():
    env = sim.Environment()
    class X(sim.Component):
        def process(self):
            with pytest.raises(ValueError):
                self.from_store(())

    store0 = sim.Store("store")
    X()
    env.run()

if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import salabim as sim
import pytest
import datetime

import collections


def collect(dis, n=10000):
    m = sim.Monitor()
    for _ in range(n):
        m.tally(dis())
    return m


def test_time_unit():
    env = sim.Environment()
    with pytest.raises(AttributeError):
        env.hours(1)
    env = sim.Environment()
    with pytest.raises(AttributeError):
        env.years(1)
    env = sim.Environment(time_unit="days")
    assert env.years(1 / 365) == pytest.approx(1)
    assert env.weeks(1 / 7) == pytest.approx(1)
    assert env.days(1) == pytest.approx(1)
    assert env.hours(24) == pytest.approx(1)
    assert env.minutes(24 * 60) == pytest.approx(1)
    assert env.seconds(24 * 60 * 60) == pytest.approx(1)
    assert env.milliseconds(24 * 60 * 60 * 1000) == pytest.approx(1)
    assert env.microseconds(24 * 60 * 60 * 1e6) == pytest.approx(1)

    assert env.to_years(365) == pytest.approx(1)
    assert env.to_weeks(7) == pytest.approx(1)
    assert env.to_days(1) == pytest.approx(1)
    assert env.to_hours(1 / 24) == pytest.approx(1)
    assert env.to_minutes(1 / (24 * 60)) == pytest.approx(1)
    assert env.to_seconds(1 / (24 * 60 * 60)) == pytest.approx(1)
    assert env.to_milliseconds(1 / (24 * 60 * 60 * 1000)) == pytest.approx(1)
    assert env.to_microseconds(1 / (24 * 60 * 60 * 1e6)) == pytest.approx(1)
    assert env.to_time_unit("days", 1) == pytest.approx(1)

    m = collect(sim.Uniform(0, 48, time_unit="hours"))
    assert m.mean() == pytest.approx(1, rel=1e-2)


def test_datetime0t():
    class X(sim.Component):
        ...

    env = sim.Environment()
    assert env.datetime0() == False
    assert env.time_to_str(1) == "     1.000"
    assert env.duration_to_str(1) == "1.000"

    env = sim.Environment(datetime0=False)
    assert env.datetime0() == False
    assert env.time_to_str(1) == "     1.000"
    assert env.duration_to_str(1) == "1.000"

    env = sim.Environment(datetime0=None)
    assert env.datetime0() == False
    assert env.time_to_str(1) == "     1.000"
    assert env.duration_to_str(1) == "1.000"

    with pytest.raises(ValueError):
        env = sim.Environment(datetime0=12)

    env = sim.Environment(datetime0=True)
    assert env.datetime0() == datetime.datetime(1970, 1, 1)

    assert env.time_to_str(1) == "Thu 1970-01-01 00:00:01"
    assert env.duration_to_str(1) == "00:00:01"

    env = sim.Environment(datetime0=datetime.datetime(2022, 4, 29))
    assert env.datetime0() == datetime.datetime(2022, 4, 29)

    assert env.time_to_str(1) == "Fri 2022-04-29 00:00:01"
    assert env.duration_to_str(1) == "00:00:01"

    env = sim.Environment(datetime0=datetime.datetime(2022, 4, 29), time_unit="days")
    assert env.datetime0() == datetime.datetime(2022, 4, 29)

    assert env.time_to_str(1) == "Sat 2022-04-30 00:00:00"
    assert env.duration_to_str(1) == "1 00:00:00"

    env = sim.Environment()
    assert env.get_time_unit() == "n/a"
    env.datetime0(True) == datetime.datetime(1970, 1, 1)
    assert env.get_time_unit() == "seconds"

    env = sim.Environment(time_unit="minutes")
    assert env.get_time_unit() == "minutes"
    env.datetime0(True) == datetime.datetime(1970, 1, 1)
    assert env.get_time_unit() == "minutes"
    env.datetime0(datetime.datetime(2022, 4, 29)) == datetime.datetime(2022, 4, 29)

    env = sim.Environment(datetime0=datetime.datetime(2022, 4, 29), time_unit="days")
    assert env.t_to_datetime(0.5) == datetime.datetime(2022, 4, 29, 12, 0)
    assert env.duration_to_timedelta(0.5) == datetime.timedelta(hours=12)
    assert env.datetime_to_t(datetime.datetime(2022, 4, 28)) == -1
    assert env.timedelta_to_duration(datetime.timedelta(hours=-12)) == -0.5

    env = sim.Environment(datetime0=True)
    assert env.datetime0() == datetime.datetime(1970, 1, 1)
    x0 = X()
    assert x0.creation_time() == 0
    assert env.time_to_str(x0.creation_time()) == "Thu 1970-01-01 00:00:00"
    env.run(3600)
    env.reset_now()
    assert x0.creation_time() == -3600
    assert env.datetime0() == datetime.datetime(1970, 1, 1, 1)
    assert env.time_to_str(x0.creation_time()) == "Thu 1970-01-01 00:00:00"
    x1 = X()
    assert x1.creation_time() == 0
    assert env.time_to_str(x1.creation_time()) == "Thu 1970-01-01 01:00:00"
    env.datetime0(datetime.datetime(2022, 4, 29))
    assert env.time_to_str(x0.creation_time()) == "Thu 2022-04-28 23:00:00"
    assert env.time_to_str(x1.creation_time()) == "Fri 2022-04-29 00:00:00"


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import salabim as sim
import pytest
import tempfile
import pickle
from array import array
import math
import datetime


def compare_output(out0, out1):
    lines0 = out0.split("\n")
    lines1 = out1.split("\n")
    if len(lines0) != len(lines1):
        return False

    for line0, line1 in zip(lines0, lines1):
        if line0.rstrip() != line1.rstrip():
            return False
    return True


def test_monitor1():
    class X(sim.Component):
        def setup(self, stats_only=False):
            self.ml = sim.Monitor("ml", level=True, stats_only=stats_only)
            self.m = sim.Monitor("m", stats_only=stats_only)

        def process(self):
            for v in range(0, 101):
                self.ml.tally(v)
                self.m.tally(v)
                self.hold(1)
                env.ended = True

    class Controller(sim.Component):
        def process(self):
            while not env.ended:
                self.hold(sim.Uniform(10, 20)())
                for i in (False, True):
                    env.x[i].ml.monitor(False)
                    env.x[i].interrupt()
                self.hold(sim.Uniform(1000, 2000)())
                for i in (False, True):
                    env.x[i].ml.monitor(True)
                    env.x[i].resume()

    env = sim.Environment()
    env.ml = sim.Monitor("ml", level=True)
    env.m = sim.Monitor("m")
    env.ended = False
    env.x = {}
    for i in (False, True):
        env.x[i] = X(stats_only=i)
    Controller()
    env.run()
    assert env.x[False].ml.mean() == pytest.approx(50)
    assert env.x[False].ml.percentile(95) == 95
    assert env.x[False].ml.maximum() == 100
    assert env.x[False].ml.minimum() == 0
    assert env.x[False].ml.duration() == pytest.approx(101)
    assert env.x[False].ml.duration(ex0=True) == pytest.approx(100)
    assert env.x[False].ml.duration_zero() == pytest.approx(1)

    assert env.x[False].m.mean() == 50
    assert env.x[False].m.percentile(95) == 95
    assert env.x[False].m.maximum() == 100
    assert env.x[False].m.minimum() == 0
    assert env.x[False].m.number_of_entries() == 101
    assert env.x[False].m.number_of_entries(ex0=True) == 100
    assert env.x[False].m.number_of_entries_zero() == 1

    assert env.x[True].ml.mean() == pytest.approx(50)
    assert env.x[True].ml.maximum() == 100
    with pytest.raises(NotImplementedError):
        env.x[True].ml.percentile(95)
    assert env.x[True].ml.minimum() == 0
    assert env.x[True].ml.duration() == pytest.approx(101)
    assert env.x[True].ml.duration(ex0=True) == pytest.approx(100)
    assert env.x[True].ml.duration_zero() == pytest.approx(1)

    assert env.x[True].m.mean() == 50
    assert env.x[True].m.maximum() == 100
    with pytest.raises(NotImplementedError):
        env.x[True].m.percentile(95)
    assert env.x[True].m.minimum() == 0
    assert env.x[True].m.number_of_entries() == 101
    assert env.x[True].m.number_of_entries(ex0=True) == 100
    assert env.x[True].m.number_of_entries_zero() == 1

    assert env.x[False].m.std() == pytest.approx(env.x[True].m.std())
    assert env.x[False].ml.std() == pytest.approx(env.x[True].ml.std())


def test_monitor2():
    env = sim.Environment()
    env.m = sim.Monitor("m")
    env.m.tally(0)
    env.m.tally(0)
    env.m.tally(1, 2)

    assert env.m.mean() == 0.5
    assert env.m.percentile(50) == pytest.approx(0.5)
    assert env.m.percentile(50, interpolation="lower") == pytest.approx(0)
    assert env.m.percentile(50, interpolation="higher") == pytest.approx(1)

    assert env.m.percentile(95) == pytest.approx(1)

    assert env.m.maximum() == 1
    assert env.m.minimum() == 0

    env.m = sim.Monitor("m")
    env.m.tally(0)
    env.m.tally(0)
    env.m.tally(1)
    env.m.tally(1)

    assert env.m.mean() == 0.5
    assert env.m.percentile(50) == pytest.approx(0.5)
    assert env.m.percentile(50, interpolation="lower") == pytest.approx(0)
    assert env.m.percentile(50, interpolation="higher") == pytest.approx(1)

    assert env.m.percentile(95) == pytest.approx(1)

    assert env.m.maximum() == 1
    assert env.m.minimum() == 0

    env.m = sim.Monitor("m")
    env.m.tally(0)
    env.m.tally(1)

    assert env.m.mean() == 0.5
    assert env.m.percentile(95) == pytest.approx(0.95)
    assert env.m.maximum() == 1
    assert env.m.minimum() == 0
    env.m = sim.Monitor("m")

    env.m.tally("red")
    env.m.tally("blue")
    env.m.tally("blue")
    env.m.tally("green")
    env.m.tally("green")
    env.m.tally("green")
    env.m.tally("green")
    env.m.tally("green")

    env.m.tally("yellow")
    env.m.tally(2)
    env.m.tally(8)

    assert env.m.mean() == pytest.approx(10 / 11)
    assert env.m.percentile(95) == pytest.approx(5)
    assert env.m.maximum() == 8
    assert env.m.minimum() == 0

    assert env.m.value_weight("blue") == 2
    assert env.m.value_weight(("blue", "green")) == 7
    assert env.m.value_weight("purple") == 0

    assert env.m.value_number_of_entries("blue") == 2
    assert env.m.value_number_of_entries(("blue", "green")) == 7
    assert env.m.value_number_of_entries("purple") == 0


def test_monitor_3():
    env = sim.Environment()
    for i in (False, True):
        m = sim.Monitor("m", level=False, stats_only=True)
        m.tally(1)
        m.tally(2)
        m.tally("3")
        m.tally("?")
        m.tally(None)
        assert m.mean() == 1.2
        assert m.number_of_entries() == 5
        assert m.number_of_entries(ex0=True) == 3
        assert m.number_of_entries_zero() == 2


def test_monitor4():
    class X(sim.Component):
        def setup(self, stats_only=False):
            self.ml = sim.Monitor("ml", level=True, initial_tally=10, stats_only=stats_only)

        def process(self):
            self.hold(till=5)
            self.ml.tally(5)
            self.hold(till=6)
            self.ml.monitor(False)
            self.ml.tally(6)
            self.hold(till=7)
            self.ml.tally(7)
            self.hold(till=7.5)
            assert self.ml() == 7
            self.hold(till=8)
            self.ml.tally(8)
            self.ml.monitor(True)
            self.hold(till=9)
            self.ml.tally(9)
            self.hold(1000)

    env = sim.Environment()
    env.ml = sim.Monitor("ml", level=True)
    env.x = {}
    for i in (False, True):
        env.x[i] = X(stats_only=i)
    env.run()
    assert env.x[False].ml.get() == 9
    assert env.x[False].ml() == 9
    assert env.x[False].ml(env.now()) == 9
    assert env.x[False].ml(0.5) == 10
    assert env.x[False].ml(8.5) == 8
    assert env.x[False].ml(6.5) == -sim.inf

    assert env.x[True].ml.get() == 9
    assert env.x[True].ml() == 9
    assert env.x[True].ml(env.now()) == 9
    with pytest.raises(NotImplementedError):
        env.x[True].ml(0.5)


def test_monitor_5():
    env = sim.Environment()
    r = sim.Resource()
    s = sim.State()
    q = sim.Queue()
    all_monitors = r.all_monitors() + s.all_monitors() + q.all_monitors()
    assert len(all_monitors) == 8 + 3 + 2
    for m in all_monitors:
        assert not m.stats_only()
    r.reset_monitors(stats_only=True)
    s.reset_monitors(stats_only=True)
    q.reset_monitors(stats_only=True)

    for m in all_monitors:
        assert m.stats_only()


def test_percentile():
    class X(sim.Component):
        def process(self):
            for value in (4, 1, 1, 3, 6, 3, 0, 7, 2, 2):
                env.ml.tally(value)
                env.m.tally(value)
                self.hold(1)

    env = sim.Environment()

    env.m = sim.Monitor()
    assert math.isnan(env.m.percentile(0))
    env.ml = sim.Monitor(level=True)
    assert env.ml.percentile(0) == 0
    env.ml = sim.Monitor(level=True, initial_tally=7)
    assert env.ml.percentile(0) == 7
    env.m.tally(1)
    assert env.m.percentile(0) == env.m.percentile(50) == env.m.percentile(100) == env.m.percentile(-1) == env.m.percentile(101) == 1
    env.ml.tally(1)
    assert env.ml.percentile(0) == env.ml.percentile(50) == env.ml.percentile(100) == env.ml.percentile(-1) == env.ml.percentile(101) == 1

    env.m = sim.Monitor()
    env.ml = sim.Monitor(level=True)
    X()
    env.run(10)
    env.m.tally(15)
    assert env.m.mean() == 4
    assert env.ml.mean() == 2.9
    assert env.m.percentile(10) == 1
    assert env.m.percentile(5) == 0.5
    assert env.m.percentile(5, interpolation="lower") == 0
    assert env.m.percentile(5, interpolation="higher") == 1
    assert env.m.percentile(5, interpolation="midpoint") == 0.5
    assert env.m.percentile(5, interpolation="linear") == 0.5
    assert env.m.percentile(5, interpolation="nearest") == 0
    assert env.m.percentile(4, interpolation="nearest") == 0
    assert env.m.percentile(6, interpolation="nearest") == 1
    with pytest.raises(ValueError):
        env.m.percentile(5, interpolation="closest") == 1

    assert env.ml.percentile(5) == 0
    assert env.ml.percentile(15) == 1
    assert env.ml.percentile(95) == 7
    assert env.ml.percentile(10) == 0.5
    assert env.ml.percentile(10, interpolation="lower") == 0
    assert env.ml.percentile(10, interpolation="higher") == 1
    assert env.ml.percentile(10, interpolation="midpoint") == 0.5
    assert env.ml.percentile(10, interpolation="linear") == 0.5
    with pytest.raises(ValueError):
        env.ml.percentile(10, interpolation="nearest") == 1


def test_pickle():
    class X(sim.Component):
        def process(self):
            for v in range(101):
                env.ml.tally(v)
                env.m.tally(v)
                self.hold(1)
                env.ended = True

    class Controller(sim.Component):
        def process(self):
            while not env.ended:
                self.hold(sim.Uniform(10, 20)())
                env.ml.monitor(False)
                env.x.interrupt()
                self.hold(sim.Uniform(1000, 2000)())
                env.ml.monitor(True)
                env.x.resume()

    env = sim.Environment()
    env.ml = sim.Monitor("ml", level=True)
    env.m = sim.Monitor("m")
    env.ended = False

    env.x = X()
    Controller()
    env.run()
    with tempfile.TemporaryDirectory() as tmp_path:
        pickle.dump(env.ml.freeze(), open(tmp_path + "ml.pickle", "wb"))
        pickle.dump(env.m.freeze(), open(tmp_path + "m.pickle", "wb"))

        env = sim.Environment()  # to reset env
        env.ml = pickle.load(open(tmp_path + "ml.pickle", "rb"))
        env.m = pickle.load(open(tmp_path + "m.pickle", "rb"))
    assert env.ml.mean() == pytest.approx(50)
    assert env.ml.percentile(95) == 95
    assert env.ml.maximum() == 100

    assert env.m.mean() == pytest.approx(50)
    assert env.m.median() == pytest.approx(50)
    assert env.m.percentile(95) == 95
    assert env.m.maximum() == 100
    assert env.m.minimum() == 0


def test_values():
    env = sim.Environment()
    env.m = sim.Monitor("m")
    values = [1, 2, 3, "one", "two", "three", 1, 1, 1, 1.1]
    for value in values:
        env.m.tally(value)
    assert compare_output(
        env.m.print_histogram(as_str=True),
        """\
Histogram of m
                        all    excl.zero         zero
-------------- ------------ ------------ ------------
entries              10            7            3
mean                  1.010        1.443
std.deviation         0.895        0.721

minimum               0            1
median                1            1
90% percentile        2.100        2.400
95% percentile        2.550        2.700
maximum               3            3

           <=       entries     %  cum%
        0             3      30    30   ************************|
        1             4      40    70   ********************************                        |
        2             2      20    90.0 ****************                                                        |
        3             1      10   100.0 ********                                                                       |
          inf         0       0   100.0                                                                                |
""",
    )

    assert compare_output(
        env.m.print_histogram(values=True, as_str=True),
        """\
Histogram of m
entries             10

value               entries     %
1                         4  40   ********************************
1.1                       1  10   ********
2                         1  10   ********
3                         1  10   ********
one                       1  10   ********
three                     1  10   ********
two                       1  10   ********
""",
    )
    set_values_in = set(values)
    set_values_out = set(env.m.values())
    assert set_values_in == set_values_out


def test_level_values():
    class X(sim.Component):
        def process(self):
            self.set_mode("abc")
            self.hold(0)
            self.request(r, mode="def")
            self.hold(10.1, mode="ghi")
            self.release()
            self.passivate(mode="jkl")

    class Interrupter(sim.Component):
        def process(self):
            self.hold(6)
            for x in (x0,x1,x2):
                if x.isscheduled():
                    x.interrupt()
            self.hold(4)
            for x in (x0,x1,x2):
                if x.isinterrupted():
                    x.resume()

    def status_map(status):
        return sim.statuses().index(status)

    modes = "abc def ghi jkl".split()

    env = sim.Environment(trace=False)

    r = sim.Resource("r")
    x0 = X(process="", mode="abc")
    x1 = X(process="")
    x2 = X(process="")
    Interrupter()

    env.run(4.3)
    x0.activate()
    x1.activate()
    x2.activate()

    env.run(till=50)

    assert x1.status.value_duration("requesting") == pytest.approx(14.1)

    assert x0.mode.xt(force_numeric=None) == (["abc", "ghi", "jkl", "jkl"], array("d", [0.0, 4.3, 18.4, 50.0]))
    assert x0.status.xt(force_numeric=None) == (
        ["data", "scheduled", "interrupted", "scheduled", "passive", "passive"],
        array("d", [0.0, 4.3, 6.0, 10.0, 18.4, 50.0]),
    )
    assert x0.status.values() == ["data", "interrupted", "passive", "scheduled"]
    assert x0.mode.values() == ["abc", "ghi", "jkl"]

    assert compare_output(
        x0.mode.print_histogram(values=True, ex0=True, as_str=True),
        """\
Histogram of x.0.mode[ex0]
duration            50

value                     duration     %
abc                          4.300   8.6 ******
ghi                         14.100  28.2 **********************
jkl                         31.600  63.2 **************************************************
""",
    )
    assert compare_output(
        x0.status.print_histogram(values=True, as_str=True),
        """\
Histogram of x.0.status
duration            50

value                     duration     %
data                         4.300   8.6 ******
interrupted                  4       8   ******
passive                     31.600  63.2 **************************************************
scheduled                   10.100  20.2 ****************
""",
    )
    assert compare_output(
        x0.status.print_histogram(values=["scheduled", "current", "passive"], as_str=True),
        """\
Histogram of x.0.status
duration            50

value                     duration     %
scheduled                   10.100  20.2 ****************
current                      0       0
passive                     31.600  63.2 **************************************************
<rest>                       8.300  16.6 *************
""",
    )
    assert compare_output(
        x0.status.print_histogram(values=["scheduled", "current", "passive"], as_str=True, sort_on_duration=True),
        """\
Histogram of x.0.status
duration            50

value                     duration     %
passive                     31.600  63.2 **************************************************
scheduled                   10.100  20.2 ****************
current                      0       0
<rest>                       8.300  16.6 *************
""",
    )

    assert compare_output(
        x0.status.print_histogram(values=True, as_str=True),
        """\
Histogram of x.0.status
duration            50    

value                     duration     %
data                         4.300   8.6 ******
interrupted                  4       8   ******
passive                     31.600  63.2 **************************************************
scheduled                   10.100  20.2 ****************
""",
    )

    assert compare_output(
        x0.status.print_histogram(values=True, sort_on_duration=True, as_str=True),
        """\
Histogram of x.0.status
duration            50    

value                     duration     %
passive                     31.600  63.2 **************************************************
scheduled                   10.100  20.2 ****************
data                         4.300   8.6 ******
interrupted                  4       8   ******
""",
    )


def test_map():
    class X1(sim.Component):
        def process(self):
            m1.tally("x1.0")
            self.hold(2)
            m1.tally("x1.2")
            self.hold(2)
            m1.tally("x1.4")

    class X2(sim.Component):
        def process(self):
            self.hold(1)
            m2.tally("x2.1")
            self.hold(1)
            m2.tally("x2.3")
            self.hold(1)
            m2.tally("x2.4")

    class M2Spoiler(sim.Component):
        def process(self):
            self.hold(1.5)
            m2.monitor(False)
            self.hold(2)
            m2.monitor(True)

    inf = sim.inf
    env = sim.Environment()
    X1()
    X2()
    M2Spoiler()
    m = sim.Monitor("m", level=False)
    m.tally(45.6)
    m.tally(6)
    m.tally(5.8)
    m1 = sim.Monitor("m1", level=True, initial_tally="x1.0")
    m2 = sim.Monitor("m2", level=True, initial_tally="x2.0")
    env.run(10)
    m3 = m1.x_map(lambda x1, x2: f"({x1},{x2})", [m2])
    assert m3.xt(force_numeric=False) == (
        ["(x1.0,x2.0)", "(x1.0,x2.1)", -inf, -inf, "(x1.2,x2.4)", "(x1.4,x2.4)", -inf, -inf],
        array("d", [0.0, 1.0, 1.5, 2.0, 3.5, 4.0, 10.0, 10.0]),
    )
    assert m.x_map(int).x() == [45, 6, 5]
    
def test_as_dataframe():
    try:
        import pandas as pd
    except ImportError:
        pytest.skip("could not import pandas")

    env = sim.Environment()
    level_monitor0 = env.Monitor("level monitor0", level=True)
    level_monitor1 = env.Monitor("level monitor1", level=True)
    non_level_monitor = env.Monitor("non level monitor")
    
    env.run(till=1)
    level_monitor0.tally(11)
    non_level_monitor.tally(11)
    env.run(till=2)
    level_monitor0.tally(12)
    non_level_monitor.tally(12)
    env.run(till=3)
    level_monitor1.tally(13)
    non_level_monitor.tally(13)
    env.run(till=4.5)
    level_monitor0.tally(14.5)
    level_monitor1.tally(14.5)
    non_level_monitor.tally(14.5)
    env.run(till=5)
    level_monitor0.tally(15)
    non_level_monitor.tally(15)
    env.run(till=10)
    level_monitor0.tally(20)
    non_level_monitor.tally(20)
    
    df = level_monitor0.as_resampled_dataframe(delta_t=1, extra_monitors=[level_monitor1], min_t=-1, max_t=22)
    df0 = pd.DataFrame(
        {
            "t": list(sim.arange(-1, 22)),
            "level monitor0.x": [pd.NA, 0, 11, 12, 12, 12, 15, 15, 15, 15, 15, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20],
            "level monitor1.x": [pd.NA, 0, 0, 0, 13, 13, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5, 14.5],
        }
    )
    assert df.equals(df0)
    
    df = non_level_monitor.as_dataframe(include_t=False)
    df0 = pd.DataFrame({"non level monitor.x": [11, 12, 13, 14.5, 15, 20]})
    assert df.equals(df0)
    
    df = non_level_monitor.as_dataframe()
    df0 = pd.DataFrame({"t": [1, 2, 3, 4.5, 5, 10], "non level monitor.x": [11, 12, 13, 14.5, 15, 20]})
    assert df.equals(df0)    

def test_as_dataframe_use_datetime0():
    try:
        import pandas as pd
    except ImportError:
        pytest.skip("could not import pandas")

    env = sim.Environment(datetime0=datetime.datetime(2023,1,1), time_unit="days")
    level_monitor0 = env.Monitor("level monitor0", level=True)
    level_monitor1 = env.Monitor("level monitor1", level=True)
    non_level_monitor = env.Monitor("non level monitor")

    env.run(till=1.3)
    level_monitor0.tally(11)
    env.run(till=2.5)
    level_monitor0.tally(12)
    env.run(till=3.1)
    level_monitor1.tally(14)
    env.run(till=4.8)
    level_monitor0.tally(14.5)
    env.run(till=5)
    env.datetime0(datetime.datetime(2023,1,1))
    df = level_monitor0.as_resampled_dataframe(delta_t=datetime.timedelta(days=2), min_t=datetime.datetime(2023,1,3), use_datetime0=True)
    d=datetime.datetime
    df0=pd.DataFrame({"t": [d(2023,1,3),d(2023,1,5)], "level monitor0.x":[11,12]})
    assert df0.equals(df)
    df = level_monitor0.as_dataframe(use_datetime0=True)
    df0=pd.DataFrame({"t":[d(2023,1,1), d(2023,1,2,7,12),d(2023,1,3,12),d(2023,1,5,19,12),d(2023,1,6,0,0)], "level monitor0.x": [0,11,12,14.5,14.5]})
    assert df0.equals(df)
    df = level_monitor0.as_dataframe()
    df0=pd.DataFrame({"t":[0,1.3,2.5,4.8,5.0], "level monitor0.x": [0,11,12,14.5,14.5]})
    assert df0.equals(df)    
    df = level_monitor0.as_dataframe(add_now=False)
    df0=pd.DataFrame({"t":[0,1.3,2.5,4.8], "level monitor0.x": [0,11,12,14.5]})
    assert df0.equals(df)    

if __name__ == "__main__":
    pytest.main(["-vv", "-s", "-x", __file__])
